import os
from typing import Type, Optional
from ulauncher.utils.mypy_extensions import TypedDict

from ulauncher.config import CONFIG_DIR
from ulauncher.utils.db.KeyValueJsonDb import KeyValueJsonDb
from ulauncher.utils.decorator.singleton import singleton

ExtensionRecord = TypedDict('ExtensionRecord', {
    'id': str,
    'url': str,
    'updated_at': str,
    'last_commit': str,
    'last_commit_time': str
})


class ExtensionDb(KeyValueJsonDb[str, ExtensionRecord]):

    @classmethod
    @singleton
    def get_instance(cls: Type['ExtensionDb']) -> 'ExtensionDb':
        db_path = os.path.join(CONFIG_DIR, 'extensions.json')
        db = cls(db_path)
        db.open()

        return db

    def find_by_url(self, url: str) -> Optional[ExtensionRecord]:
        for ext in self.get_records().values():
            if ext['url'] == url:
                return ext

        return None



from functools import wraps
from gi.repository import GLib


def glib_idle_add(fn):
    """
    Schedules fn to run in the main GTK loop
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        GLib.idle_add(fn, *args, **kwargs)

    wrapper.original = fn

    return wrapper



from distutils.version import StrictVersion
from gi.repository import Gtk


def gtk_version_is_gte(major: int, minor: int, micro: int) -> bool:
    gtk_version = '%s.%s.%s' % (Gtk.get_major_version(), Gtk.get_minor_version(), Gtk.get_micro_version())
    return StrictVersion(gtk_version) >= StrictVersion('%s.%s.%s' % (major, minor, micro))



import os
from ulauncher.config import CACHE_DIR
from ulauncher.utils.db.KeyValueDb import KeyValueDb
from ulauncher.utils.decorator.singleton import singleton


class AppCacheDb(KeyValueDb):

    @classmethod
    @singleton
    def get_instance(cls):
        db = cls(os.path.join(CACHE_DIR, 'app_cache.db'))
        db.open()
        return db



import os
import logging
from time import time, sleep
from functools import wraps
from typing import Dict

import pyinotify

from ulauncher.utils.decorator.run_async import run_async
from ulauncher.utils.desktop.reader import find_desktop_files, read_desktop_file, filter_app, find_apps_cached
from ulauncher.search.apps.AppDb import AppDb
from ulauncher.config import DESKTOP_DIRS

logger = logging.getLogger(__name__)


def _only_desktop_files(func):
    """
    Decorator for pyinotify.ProcessEvent
    Triggers event handler only for *.desktop files
    """

    @wraps(func)
    def decorator_func(self, event, *args, **kwargs):
        if os.path.splitext(event.pathname)[1] == '.desktop':
            return func(self, event, *args, **kwargs)
        return None

    return decorator_func


DeferredFiles = Dict[str, float]


class AppNotifyEventHandler(pyinotify.ProcessEvent):
    RETRY_INTERVAL = 2  # seconds
    RETRY_TIME_SPAN = (5, 30)  # make an attempt to process desktop file within 5 to 30 seconds after event came in
    # otherwise application icon or .desktop file itself may not be ready

    class InvalidDesktopFile(IOError):
        pass

    class SkipDesktopFile(Exception):
        pass

    def __init__(self, db):
        super(AppNotifyEventHandler, self).__init__()
        self.__db = db  # type: AppDb

        # key is a file path, value is an addition time
        self._deferred_files = {}  # type: DeferredFiles
        self._init_worker()

    @run_async(daemon=True)
    def _init_worker(self) -> None:
        """
        Add files to the DB with some delay,
        otherwise .desktop file may not be ready while application is being installed
        """
        while True:
            for pathname, start_time in list(self._deferred_files.items()):
                time_passed = time() - start_time
                if time_passed < self.RETRY_TIME_SPAN[0]:
                    # skip this file for now
                    continue

                if time_passed > self.RETRY_TIME_SPAN[1]:
                    # give up on file after time limit
                    if pathname in self._deferred_files:
                        del self._deferred_files[pathname]

                try:
                    self._add_file_sync(pathname)
                except self.InvalidDesktopFile:
                    # retry
                    pass
                except self.SkipDesktopFile:
                    # skip adding, and break the loop
                    if pathname in self._deferred_files:
                        del self._deferred_files[pathname]
                    break
                # pylint: disable=broad-except
                except Exception as e:
                    # give up on unexpected exception
                    logger.warning("Unexpected exception: %s", e)
                    if pathname in self._deferred_files:
                        del self._deferred_files[pathname]
                else:
                    # success
                    if pathname in self._deferred_files:
                        del self._deferred_files[pathname]

            sleep(self.RETRY_INTERVAL)

    def add_file_deferred(self, pathname: str) -> None:
        """
        Add .desktop file to DB a little bit later
        """
        self._deferred_files[pathname] = time()

    def _add_file_sync(self, pathname: str) -> None:
        """
        Add .desktop file to DB

        Raises self.InvalidDesktopFile if failed to add an app
        """

        # get filename of the desktop file (i.e chromium.desktop)
        file_name = os.path.basename(pathname)
        # search for desktop file in all of DESKTOP_DIRS
        pathnames_in_xdg_dirs = list(find_desktop_files(DESKTOP_DIRS, file_name))
        # if the pathname is not found in the desktop files it is overridden
        # with a file of the same name in a different XDG directory, so skip
        # trying to add this desktop file
        if pathname not in pathnames_in_xdg_dirs:
            logger.warning('Skipping adding %s to DB -> desktop file overridden in a different XDG directory', pathname)
            raise self.SkipDesktopFile(pathname)

        try:
            app = read_desktop_file(pathname)
            if filter_app(app):
                self.__db.put_app(app)
                logger.info('New app was added "%s" (%s)', app.get_name(), app.get_filename())
            else:
                raise self.InvalidDesktopFile(pathname)
        except Exception as e:
            logger.warning('Cannot add %s to DB -> %s', pathname, e)
            raise self.InvalidDesktopFile(pathname)

    def _remove_file(self, pathname: str) -> None:
        """
        Remove .desktop file from DB
        :param str pathname:
        """
        self.__db.remove_by_path(pathname)
        logger.info('App was removed (%s)', pathname)

    @_only_desktop_files
    def process_IN_CREATE(self, event):
        self.add_file_deferred(event.pathname)

    @_only_desktop_files
    def process_IN_DELETE(self, event):
        self._remove_file(event.pathname)

    @_only_desktop_files
    def process_IN_MODIFY(self, event):
        self.add_file_deferred(event.pathname)

    @_only_desktop_files
    def process_IN_MOVED_FROM(self, event):
        self._remove_file(event.pathname)

    @_only_desktop_files
    def process_IN_MOVED_TO(self, event):
        self.add_file_deferred(event.pathname)


@run_async(daemon=True)
def start():
    """
    Add all known .desktop files to the DB and start inotify watcher
    """

    db = AppDb.get_instance()
    t0 = time()
    logger.info('Started scanning desktop dirs')
    for app in find_apps_cached():
        db.put_app(app)
    logger.info('Scanned desktop dirs in %.2f seconds', (time() - t0))

    wm = pyinotify.WatchManager()
    handler = AppNotifyEventHandler(db)
    notifier = pyinotify.ThreadedNotifier(wm, handler)
    notifier.setDaemon(True)
    logger.debug('Start watching desktop files...')
    notifier.start()
    # pylint: disable=no-member
    mask = pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_MODIFY | \
        pyinotify.IN_MOVED_FROM | pyinotify.IN_MOVED_TO
    wm.add_watch(DESKTOP_DIRS, mask, rec=True, auto_add=True)



# pylint: disable=too-few-public-methods
class Response:
    """
    :param ~ulauncher.api.shared.event.BaseEvent event:
    :param ~ulauncher.api.shared.action.BaseAction.BaseAction action:
    """

    event = None

    action = None

    def __init__(self, event, action):
        self.event = event
        self.action = action



import os


class DesktopParser:
    DESKTOP_SECTION = '[Desktop Entry]'

    __property_list = None

    def __init__(self, filename):
        self.__property_list = []
        self.set_filename(filename)
        self.read()
        super(DesktopParser, self).__init__()

    def set_filename(self, filename):
        self._filename = filename

    def read(self):
        """
        Read [Desktop Entry] section and save key=values pairs to __property_list
        """
        with open(self._filename, 'r') as f:
            is_desktop_section = False
            for line in f.readlines():
                line = line.strip(' ' + os.linesep)
                if line == self.DESKTOP_SECTION:
                    is_desktop_section = True
                    continue
                if line.startswith('['):
                    # another section begins
                    is_desktop_section = False
                    continue
                if is_desktop_section and '=' in line:
                    (key, value) = line.split('=', 1)
                    self.set(key.strip(), value.strip())

    def write(self):
        """
        Write properties to the file
        """
        dir = os.path.dirname(self._filename)
        if not os.path.exists(dir):
            os.makedirs(dir)

        with open(self._filename, 'w') as f:
            f.write(os.linesep.join((self.DESKTOP_SECTION, os.linesep.join(['='.join((str(k), str(v).strip()))
                                                                            for k, v in self.__property_list]))))

    def get(self, name):
        """
        Raises KeyError if name is not found
        """

        for key, value in self.__property_list:
            if key.lower() == name.lower():
                return value
        raise KeyError('%s' % name)

    def set(self, name, value):
        if not name:
            raise ValueError("Invalid value for name: '%s'" % name)

        for i, (key, _) in enumerate(self.__property_list):
            if key.lower() == name.lower():
                self.__property_list[i] = (key, value)
                return

        self.__property_list.append((name, value))

    def get_boolean(self, name):
        """
        Returns True if value is "1", "yes", "true", or "on"

        Returns False if value is "0", "no", "false", or "off"

        String values are checked in a case-insensitive manner.

        Any other value will cause it to raise ValueError.
        """

        value = self.get(name).lower()
        if value in ("1", "yes", "true", "on"):
            return True
        if value in ("0", "no", "false", "off"):
            return False

        raise ValueError("Cannot coerce '%s' to boolean" % value)



from ulauncher.api.shared.action.DoNothingAction import DoNothingAction


class BaseSearchMode:

    # pylint: disable=unused-argument
    def is_enabled(self, query):
        """
        Return True if mode should be enabled for a query
        """
        return False

    def on_query_change(self, query):
        """
        Triggered when user changes a search query
        """

    def handle_key_press_event(self, widget, event, query):
        """
        :param Gdk.Widget widget:
        :param Gdk.EventKey event:
        :param ~ulauncher.search.Query.Query query:
        :rtype: :class:`BaseAction`
        """
        return DoNothingAction()

    def handle_query(self, query):
        """
        :rtype: :class:`BaseAction`
        """
        return DoNothingAction()

    def get_default_items(self):
        """
        Returns a list of default result items that
        should be displayed if no results found
        """
        return []

    def get_searchable_items(self):
        """
        Returns a list of result items that
        can be looked up by name or keyword
        """
        return []



import os
import logging
from copy import copy
from ulauncher.config import DATA_DIR


# The background is set with 40 plus the number of the color, and the foreground with 30
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

# These are the sequences need to get colored ouput
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"


COLORS = {
    'WARNING': YELLOW,
    'INFO': WHITE,
    'DEBUG': BLUE,
    'CRITICAL': YELLOW,
    'ERROR': RED
}


class ColoredFormatter(logging.Formatter):
    def __init__(self, format, time_format=None, use_color=True):
        logging.Formatter.__init__(self, format, time_format)
        self.use_color = use_color

    def format(self, record):
        record = copy(record)
        levelname = record.levelname
        if self.use_color and levelname in COLORS:
            levelname_color = COLOR_SEQ % (30 + COLORS[levelname]) + levelname + RESET_SEQ
            record.levelname = levelname_color
        return logging.Formatter.format(self, record)


def setup_logging(opts):
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    format = "%(asctime)s | %(levelname)s | %(name)s: %(funcName)s() | %(message)s"

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG if opts.verbose else logging.WARNING)
    stream_handler.setFormatter(ColoredFormatter(format))
    root.addHandler(stream_handler)

    # set up login to a file
    log_file = os.path.join(DATA_DIR, 'last.log')
    if os.path.exists(log_file):
        os.remove(log_file)

    file_handler = logging.FileHandler(log_file)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(format))
    root.addHandler(file_handler)



import os


def find_extensions(ext_dir):
    """
    Yields `(extension_id, extension_path)` tuples found in a given extensions dir
    """
    if not os.path.exists(ext_dir):
        return

    dirs = [d for d in os.listdir(ext_dir) if os.path.isdir(os.path.join(ext_dir, d))]
    for dir in dirs:
        ext_path = os.path.join(ext_dir, dir)
        if os.path.isfile(os.path.join(ext_path, 'manifest.json')):
            yield (dir, ext_path)



class Query(str):
    """
    Parses user's query
    """

    def get_keyword(self):
        return self.strip().split(' ', 1)[0]

    def is_mode_active(self):
        """
        Mode is active when query starts with keyword + space
        """
        kw = self.get_keyword()
        return kw and self.startswith('%s ' % kw)

    def get_argument(self, default=None):
        try:
            return self.strip().split(' ', 1)[1].strip()
        except IndexError:
            return default



from ulauncher.api.shared.action.BaseAction import BaseAction


class ActionList(list, BaseAction):
    """
    Used to run multiple action at once

    :param list actions: list of actions to run
    """

    def keep_app_open(self):
        """
        :rtype: boolean
        :returns: return true if there are no actions in the list
                  otherwise returns ``any(map(lambda i: i.keep_app_open(), self))``
        """
        if not self:
            return True

        return any(map(lambda i: i.keep_app_open(), self))

    def run(self):
        for item in self:
            item.run()



import os
import json
from distutils.dir_util import mkpath

from ulauncher.utils.db.KeyValueDb import KeyValueDb, Key, Value


class KeyValueJsonDb(KeyValueDb[Key, Value]):
    """
    Key-value JSON database
    Use open() method to load DB from a file and commit() to save it
    """

    def open(self) -> 'KeyValueJsonDb':
        """Create a new data base or open existing one"""
        if os.path.exists(self._name):
            if not os.path.isfile(self._name):
                raise IOError("%s exists and is not a file" % self._name)

            with open(self._name, 'r') as _in:
                self.set_records(json.load(_in))
        else:
            # make sure path exists
            mkpath(os.path.dirname(self._name))
            self.commit()

        return self

    def commit(self) -> 'KeyValueJsonDb':
        """Write the database to a file"""
        with open(self._name, 'w') as out:
            json.dump(self._records, out, indent=4)
            out.close()

        return self




def alias(alternative_function_name):
    '''see http://www.drdobbs.com/web-development/184406073#l9'''
    def decorator(function):
        '''attach alternative_function_name(s) to function'''
        if not hasattr(function, 'aliases'):
            function.aliases = []
        function.aliases.append(alternative_function_name)
        return function
    return decorator



import os
from time import time, sleep
from typing import Optional
from ulauncher.config import CACHE_DIR
from ulauncher.utils.db.KeyValueDb import KeyValueDb
from ulauncher.utils.decorator.run_async import run_async
from ulauncher.utils.decorator.singleton import singleton


class FileQueries(KeyValueDb[str, float]):
    __last_put_time = None  # type: Optional[float]
    __last_save_time = None  # type: Optional[float]

    @classmethod
    @singleton
    def get_instance(cls) -> 'FileQueries':
        browser_cache = os.path.join(CACHE_DIR, 'file_browser_queries_v2.db')
        db = cls(browser_cache)
        db.open()
        return db

    def __init__(self, basename: str) -> None:
        super(FileQueries, self).__init__(basename)
        self._init_autosave()

    @run_async(daemon=True)
    def _init_autosave(self) -> None:
        """
        We don't want to trigger file I/O on every insert to the DB,
        so we will commit changes asynchronously every 20 sec
        """
        while True:
            if self.__last_save_time is None or self.__last_put_time is None or \
                    self.__last_save_time < self.__last_put_time:
                self.commit()
                self.__last_save_time = time()
            sleep(20)

    def save_query(self, path: str) -> None:
        self.__last_put_time = time()
        super(FileQueries, self).put(path, self.__last_put_time)



import re
import json
import logging
from datetime import datetime
from urllib.request import urlopen
from urllib.error import HTTPError
from typing import Dict, List, cast
from ulauncher.utils.mypy_extensions import TypedDict

from ulauncher.utils.date import iso_to_datetime
from ulauncher.api.version import api_version
from ulauncher.api.shared.errors import ErrorName, UlauncherAPIError
from ulauncher.utils.semver import satisfies, valid_range

DEFAULT_GITHUB_BRANCH = 'master'

logger = logging.getLogger(__name__)

ManifestPreference = TypedDict('ManifestPreference', {
    'default_value': str,
    'id': str,
    'name': str,
    'type': str
})

ManifestOptions = TypedDict('ManifestOptions', {
    'query_debounce': float
})

Manifest = TypedDict('Manifest', {
    'required_api_version': str,
    'description': str,
    'developer_name': str,
    'icon': str,
    'name': str,
    'options': ManifestOptions,
    'preferences': List[ManifestPreference]
})

Commit = TypedDict('Commit', {
    'sha': str,
    'time': datetime
})


class GithubExtensionError(UlauncherAPIError):
    pass


class GithubExtension:

    url_match_pattern = r'^https:\/\/github.com\/([\w-]+\/[\w-]+)$'
    url_file_template = 'https://raw.githubusercontent.com/{project_path}/{branch}/{file_path}'
    url_commit_template = 'https://api.github.com/repos/{project_path}/commits/{commit}'

    def __init__(self, url):
        self.url = url

    def validate_url(self):
        if not re.match(self.url_match_pattern, self.url, re.I):
            raise GithubExtensionError('Invalid GithubUrl: %s' % self.url, ErrorName.InvalidGithubUrl)

    def find_compatible_version(self) -> Commit:
        """
        Finds maximum version that is compatible with current version of Ulauncher
        and returns a commit or branch/tag name

        :raises ulauncher.api.server.GithubExtension.InvalidVersionsFileError:
        """
        sha_or_branch = ""
        for ver in self.read_versions():
            if satisfies(api_version, ver['required_api_version']):
                sha_or_branch = ver['commit']

        if not sha_or_branch:
            raise GithubExtensionError(
                "This extension is not compatible with current version Ulauncher extension API (v%s)" % api_version,
                ErrorName.IncompatibleVersion)

        return self.get_commit(sha_or_branch)

    def get_commit(self, sha_or_branch) -> Commit:
        """
        Github response example: https://api.github.com/repos/Ulauncher/Ulauncher/commits/a1304f5a
        """

        project_path = self._get_project_path()
        url = self.url_commit_template.format(project_path=project_path, commit=sha_or_branch)
        resp = json.loads(urlopen(url).read().decode('utf-8'))

        return {
            'sha': resp['sha'],
            'time': iso_to_datetime(resp['commit']['committer']['date'])
        }

    def _read_json(self, commit, file_path) -> Dict:
        project_path = self._get_project_path()
        url = self.url_file_template.format(project_path=project_path, branch=commit, file_path=file_path)

        try:
            return json.loads(urlopen(url).read().decode('utf-8'))
        except HTTPError as e:
            logger.warning('_read_json("%s", "%s") failed. %s: %s', commit, file_path, type(e).__name__, e)
            if e.code == 404:
                raise GithubExtensionError('Could not find versions.json file using URL "%s"' %
                                           url, ErrorName.VersionsJsonNotFound)
            raise GithubExtensionError('Unexpected Github API Error', ErrorName.GithubApiError)

    def read_versions(self) -> List[Dict[str, str]]:
        versions = self._read_json('master', 'versions.json')

        if not isinstance(versions, list):
            raise GithubExtensionError('versions.json should contain a list', ErrorName.InvalidVersionsJson)
        for ver in versions:
            if not isinstance(ver, dict):
                raise GithubExtensionError('versions.json should contain a list of objects',
                                           ErrorName.InvalidVersionsJson)
            if not isinstance(ver.get('required_api_version'), str):
                raise GithubExtensionError(
                    'versions.json: required_api_version should be a string', ErrorName.InvalidVersionsJson)
            if not isinstance(ver.get('commit'), str):
                raise GithubExtensionError('versions.json: commit should be a string', ErrorName.InvalidVersionsJson)

            valid = False
            try:
                valid = valid_range(ver['required_api_version'], False)
            # pylint: disable=broad-except
            except Exception:
                pass
            if not valid:
                raise GithubExtensionError("versions.json: invalid range '%s'" % ver['required_api_version'],
                                           ErrorName.InvalidVersionsJson)

        return versions

    def read_manifest(self, commit) -> Manifest:
        return cast(Manifest, self._read_json(commit, 'manifest.json'))

    def get_download_url(self, commit: str = DEFAULT_GITHUB_BRANCH) -> str:
        """
        >>> https://github.com/Ulauncher/ulauncher-timer
        <<< https://github.com/Ulauncher/ulauncher-timer/archive/master.zip
        """
        return '%s/archive/%s.zip' % (self.url, commit)

    def get_ext_id(self) -> str:
        """
        >>> https://github.com/Ulauncher/ulauncher-timer
        <<< com.github.ulauncher.ulauncher-timer
        """
        return 'com.github.%s' % self._get_project_path().replace('/', '.').lower()

    def _get_project_path(self) -> str:
        """
        >>> https://github.com/Ulauncher/ulauncher-timer
        <<< Ulauncher/ulauncher-timer
        """
        match = re.match(self.url_match_pattern, self.url, re.I)
        if not match:
            raise GithubExtensionError('Invalid GithubUrl: %s' % self.url, ErrorName.InvalidGithubUrl)

        return match.group(1)



import logging
import os
import sys
from subprocess import Popen
from typing import Dict, Optional
from time import time, sleep
from enum import Enum

from ulauncher.config import EXTENSIONS_DIR, ULAUNCHER_APP_DIR, get_options
from ulauncher.utils.decorator.run_async import run_async
from ulauncher.utils.mypy_extensions import TypedDict
from ulauncher.utils.decorator.singleton import singleton
from ulauncher.api.server.ExtensionManifest import ExtensionManifest
from ulauncher.api.server.ExtensionServer import ExtensionServer
from ulauncher.api.server.extension_finder import find_extensions

logger = logging.getLogger(__name__)

ExtRunError = TypedDict('ExtRunError', {
    'name': str,
    'message': str
})


class ExtRunErrorName(Enum):
    NoExtensionsFlag = 'NoExtensionsFlag'
    Terminated = 'Terminated'
    ExitedInstantly = 'ExitedInstantly'
    Exited = 'Exited'


class ExtensionRunner:

    @classmethod
    @singleton
    def get_instance(cls) -> 'ExtensionRunner':
        return cls(ExtensionServer.get_instance())

    def __init__(self, extension_server):
        self.extensions_dir = EXTENSIONS_DIR  # type: str
        self.extension_errors = {}  # type: Dict[str, ExtRunError]
        self.extension_procs = {}
        self.extension_server = extension_server
        self.dont_run_extensions = get_options().no_extensions
        self.verbose = get_options().verbose

    def run_all(self):
        """
        Finds all extensions in `EXTENSIONS_DIR` and runs them
        """
        for id, _ in find_extensions(self.extensions_dir):
            try:
                self.run(id)
            # pylint: disable=broad-except
            except Exception as e:
                logger.error("Couldn't run '%s'. %s: %s", id, type(e).__name__, e)

    def run(self, extension_id):
        """
        * Validates manifest
        * Runs extension in a new process

        :rtype: :class:`threading.Thread`
        """
        if self.is_running(extension_id):
            raise ExtensionIsRunningError('Extension ID: %s' % extension_id)

        manifest = ExtensionManifest.open(extension_id)
        manifest.validate()
        manifest.check_compatibility()

        run_process = run_async(daemon=True)(self._run_process)
        run_process(extension_id)

    def _run_process(self, extension_id):
        """
        Blocking function
        """
        cmd = [sys.executable, os.path.join(self.extensions_dir, extension_id, 'main.py')]
        env = os.environ.copy()
        env['ULAUNCHER_WS_API'] = self.extension_server.generate_ws_url(extension_id)
        env['PYTHONPATH'] = ':'.join(filter(bool, [ULAUNCHER_APP_DIR, os.getenv('PYTHONPATH')]))

        if self.verbose:
            env['VERBOSE'] = '1'

        if self.dont_run_extensions:
            args = [env.get('VERBOSE', ''), env['ULAUNCHER_WS_API'], env['PYTHONPATH']]
            args.extend(cmd)
            run_cmd = 'VERBOSE={} ULAUNCHER_WS_API={} PYTHONPATH={} {} {}'.format(*args)
            logger.warning('Copy and run the following command to start %s', extension_id)
            logger.warning(run_cmd)
            self.set_extension_error(extension_id, ExtRunErrorName.NoExtensionsFlag, run_cmd)

            return

        while True:
            t_start = time()
            proc = Popen(cmd, env=env)
            logger.info('Extension "%s" started. PID %s', extension_id, proc.pid)
            self.extension_procs[extension_id] = proc
            self.unset_extension_error(extension_id)
            code = proc.wait()

            if code <= 0:
                error_msg = 'Extension "%s" was terminated with code %s' % (extension_id, code)
                logger.error(error_msg)
                self.set_extension_error(extension_id, ExtRunErrorName.Terminated, error_msg)
                try:
                    del self.extension_procs[extension_id]
                except KeyError:
                    pass

                break

            if time() - t_start < 1:
                error_msg = 'Extension "%s" exited instantly with code %s' % (extension_id, code)
                logger.error(error_msg)
                self.set_extension_error(extension_id, ExtRunErrorName.ExitedInstantly, error_msg)
                try:
                    del self.extension_procs[extension_id]
                except KeyError:
                    pass

                break

            error_msg = 'Extension "%s" exited with code %s. Restarting...' % (extension_id, code)
            self.set_extension_error(extension_id, ExtRunErrorName.Exited, error_msg)
            logger.error(error_msg)

    def stop(self, extension_id):
        """
        Terminates extension
        """
        if not self.is_running(extension_id):
            raise ExtensionIsNotRunningError('Extension ID: %s' % extension_id)

        logger.info('Terminating extension "%s"', extension_id)
        proc = self.extension_procs[extension_id]
        try:
            del self.extension_procs[extension_id]
        except KeyError:
            pass

        terminate = run_async(proc.terminate)
        terminate()

        sleep(0.5)
        if proc.poll() is None:
            logger.warning("Kill extension \"%s\" since it doesn't react to SIGTERM", extension_id)
            proc.kill()

    def is_running(self, extension_id: str) -> bool:
        return extension_id in self.extension_procs.keys()

    def set_extension_error(self, extension_id: str, errorName: ExtRunErrorName, message: str):
        self.extension_errors[extension_id] = {
            'name': errorName.value,
            'message': message
        }

    def unset_extension_error(self, extension_id: str):
        try:
            del self.extension_errors[extension_id]
        except KeyError:
            pass

    def get_extension_error(self, extension_id: str) -> Optional[ExtRunError]:
        return self.extension_errors.get(extension_id)


class ExtensionIsRunningError(RuntimeError):
    pass


class ExtensionIsNotRunningError(RuntimeError):
    pass



from typing import Any, Callable, Optional

from ulauncher.api.shared.action.BaseAction import BaseAction
from ulauncher.search.Query import Query
from ulauncher.utils.text_highlighter import highlight_text
from ulauncher.utils.display import get_monitor_scale_factor

OnEnterCallback = Optional[Callable[[Query], Optional[BaseAction]]]


# pylint: disable=too-many-instance-attributes
class ResultItem:

    ICON_SIZE = 40
    UI_FILE = 'result_item'

    score = None  # used by SortedResultList class to maintain sorted by score order of items

    _name = None  # type: str
    _description = None  # type: str
    _keyword = None  # type: str
    _icon = None  # type: Any
    _include_in_results = True  # type: bool
    _selected_by_default = False  # type: bool
    _on_enter = None  # type: OnEnterCallback
    _on_alt_enter = None  # type: OnEnterCallback
    _highlightable = True  # type: bool
    _is_extension = False  # type: bool

    # pylint: disable=too-many-arguments
    def __init__(self,
                 name: str = '',
                 description: str = '',
                 keyword: str = '',
                 icon: Any = None,
                 include_in_results: bool = True,
                 selected_by_default: bool = False,
                 highlightable: bool = True,
                 on_enter: OnEnterCallback = None,
                 on_alt_enter: OnEnterCallback = None):
        if not isinstance(name, str):
            raise TypeError('"name" must be of type "str", "%s" given' % type(name).__name__)
        if not isinstance(description, str):
            raise TypeError('"description" must be of type "str", "%s" given' % type(description).__name__)
        if not isinstance(keyword, str):
            raise TypeError('"keyword" must be of type "str", "%s" given' % type(keyword).__name__)
        self._name = name
        self._description = description
        self._keyword = keyword
        self._icon = icon
        self._include_in_results = include_in_results
        self._selected_by_default = selected_by_default
        self._on_enter = on_enter
        self._on_alt_enter = on_alt_enter
        self._highlightable = highlightable

    @classmethod
    def get_icon_size(cls):
        return cls.ICON_SIZE * get_monitor_scale_factor()

    def get_keyword(self) -> str:
        """
        If keyword is defined, search will be performed by keyword, otherwise by name.
        """
        return self._keyword

    def get_name(self) -> str:
        return self._name

    def get_search_name(self) -> str:
        """
        Returns string that will be used for search
        :rtype: str
        """
        return self.get_name()

    def get_name_highlighted(self, query: Query, color: str) -> Optional[str]:
        """
        :param ~ulauncher.search.Query.Query query:
        :param str color:
        :rtype: str
        """
        if query and self._highlightable:
            return highlight_text(query if not self._is_extension else query.get_argument(''),
                                  self.get_name(),
                                  open_tag='<span foreground="%s">' % color,
                                  close_tag='</span>')
        # don't highlight if query is empty
        return self.get_name()

    # pylint: disable=unused-argument
    def get_description(self, query: Query) -> str:
        """
        optional

        :param ~ulauncher.search.Query.Query query:
        """
        return self._description

    def get_icon(self):
        """
        optional

        :rtype: :class:`Gtk.PixBuf`
        """
        return self._icon

    def include_in_results(self):
        """
        Return True to display item among apps in the default search
        """
        return self._include_in_results

    def selected_by_default(self, query):
        """
        Return True if item should be selected by default
        """
        return self._selected_by_default

    def on_enter(self, query: Query) -> Optional[BaseAction]:
        """
        :param ~ulauncher.search.Query.Query query: it is passed only if :meth:`get_keyword` is implemented.
                                                    This allows you to create flows with a result item
        :rtype: :class:`~ulauncher.api.shared.action.BaseAction.BaseAction`
        """
        return self._on_enter(query) if callable(self._on_enter) else None

    def on_alt_enter(self, query: Query) -> Optional[BaseAction]:
        """
        Optional alternative enter

        :param ~ulauncher.search.Query.Query query: it is passed only if :meth:`get_keyword` is implemented.
                                                    This allows you to create flows with a result item
        :rtype: :class:`~ulauncher.api.shared.action.BaseAction.BaseAction`
        """
        return self._on_alt_enter(query) if callable(self._on_alt_enter) else None



from gi.repository import GLib
from ulauncher.api.shared.action.BaseAction import BaseAction


class RenderResultListAction(BaseAction):
    """
    Renders list of result items

    :param list result_list: list of :class:`~ulauncher.api.shared.item.ResultItem.ResultItem` objects
    """

    def __init__(self, result_list):
        self.result_list = result_list

    def keep_app_open(self):
        return True

    def run(self):
        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow

        window = UlauncherWindow.get_instance()
        if window.is_visible():
            # update UI in the main thread to avoid race conditions
            GLib.idle_add(window.show_results, self.result_list)



from ulauncher.ui.ResultItemWidget import ResultItemWidget


class SmallResultItemWidget(ResultItemWidget):
    """
    It is instantiated automagically if the following is done:
        - its name is set in .ui file in class attribute
        - __gtype_name__ is set to the same class name
        - this class is be imported somewhere in the code before .ui file is built
    """

    __gtype_name__ = "SmallResultItemWidget"



from ulauncher.utils.fuzzy_search import get_matching_indexes


def highlight_text(query, text: str, open_tag='<span foreground="white">', close_tag='</span>') -> str:
    """
    Highlights words from query in a given text string
    :returns: string with Pango markup
    """
    positions = get_matching_indexes(query, text)

    # use positions to highlight text with tags
    hl_started = False
    hlted = []
    for i, char in enumerate(text):
        if i in positions and not hl_started:
            hl_started = True
            hlted.append(open_tag)
        elif i not in positions and hl_started:
            hl_started = False
            hlted.append(close_tag)

        hlted.append(char)

    if hl_started:
        # don't forget to close tag if it is opened
        hlted.append(close_tag)

    return ''.join(hlted)



from ulauncher.api.shared.item.SmallResultItem import SmallResultItem
from ulauncher.api.shared.item.ExtensionResultItem import ExtensionResultItem


class ExtensionSmallResultItem(ExtensionResultItem, SmallResultItem):
    pass



import os


class Path:
    # cache for better performance
    __cached_existing_dir = None
    _path = None  # type: str

    def __init__(self, path):
        self._path = os.path.expandvars(os.path.expanduser(path.lstrip()))

    def get_abs_path(self) -> str:
        return self._path

    def exists(self) -> bool:
        return os.path.exists(self._path)

    def get_basename(self) -> str:
        return os.path.basename(self._path.rstrip('/'))

    def get_dirname(self) -> str:
        return os.path.dirname(self.get_user_path())

    def is_dir(self) -> bool:
        return os.path.isdir(self._path)

    def is_exe(self) -> bool:
        return os.access(self._path, os.X_OK)

    def get_ext(self) -> str:
        return os.path.splitext(self._path)[1].lower()[1:]

    def get_user_path(self) -> str:
        result = self._path.rstrip('/')

        user_dir = os.path.expanduser('~')
        if result.startswith(user_dir):
            return result.replace(user_dir, '~', 1)

        return result

    def get_existing_dir(self) -> str:
        """
        Example (assuming foo & bar do not exist):

        >>> p = Path('/home/aleksandr/projects/foo/bar')
        >>> p.get_existing_dir() == '/home/aleksandr/projects'

        :returns: path to the last dir that exists in the user's query
        """
        if self.__cached_existing_dir:
            return self.__cached_existing_dir

        result = self._path
        while result and (not os.path.exists(result) or not os.path.isdir(result)):
            result = os.path.dirname(result)

        # special case when dir ends with "/."
        # we want to return path without .
        # example: /bin/env/. -> /bin/env
        if result.endswith('/.'):
            result = result[:-2]

        if not result:
            raise InvalidPathError('Invalid path "%s"' % self._path)

        self.__cached_existing_dir = result
        return result

    def get_search_part(self) -> str:
        """
        :returns: remaining part of the query that goes after :meth:`get_existing_dir`
        """
        return self._path[len(self.get_existing_dir()):].strip('/')


class InvalidPathError(RuntimeError):
    pass



import logging
import pipes
import subprocess

from ulauncher.utils.desktop.reader import read_desktop_file
from ulauncher.utils.Settings import Settings
from ulauncher.api.shared.action.BaseAction import BaseAction

logger = logging.getLogger(__name__)
settings = Settings.get_instance()


class LaunchAppAction(BaseAction):
    """
    Launches app by given `.desktop` file path

    :param str filename: path to .desktop file
    """

    def __init__(self, filename):
        self.filename = filename

    def keep_app_open(self):
        return False

    def run(self):
        app = read_desktop_file(self.filename)
        command = app.get_string('Exec')
        terminal_exec = settings.get_property('terminal-command')
        if app.get_boolean('Terminal') and terminal_exec and command:
            logger.info('Run command %s (%s) in preferred terminal (%s)', command, self.filename, terminal_exec)
            subprocess.Popen(terminal_exec.split() + [pipes.quote(command)])
        else:
            logger.info('Run application %s (%s)', app.get_name(), self.filename)
            try:
                app.launch()
            except Exception as e:
                logger.error('%s: %s', type(e).__name__, e)



from ulauncher.api.shared.action.OpenAction import OpenAction
from ulauncher.api.shared.item.SmallResultItem import SmallResultItem
from ulauncher.utils.image_loader import get_themed_icon_by_name


class OpenFolderItem(SmallResultItem):
    """
    :param ~ulauncher.utils.Path.Path path:
    """

    # pylint: disable=super-init-not-called
    def __init__(self, path):
        self.path = path
        self._name = 'Open Folder'

    def set_name(self, name):
        self._name = name

    # pylint: disable=arguments-differ
    def get_name_highlighted(self, *args):
        pass

    def get_name(self):
        return self._name

    def get_icon(self):
        return get_themed_icon_by_name('system-file-manager', self.get_icon_size())

    def on_enter(self, query):
        return OpenAction(self.path.get_abs_path())



import re

from ulauncher.api.shared.action.ActionList import ActionList
from ulauncher.api.shared.action.OpenUrlAction import OpenUrlAction
from ulauncher.api.shared.action.RunScriptAction import RunScriptAction
from ulauncher.api.shared.action.SetUserQueryAction import SetUserQueryAction
from ulauncher.api.shared.item.ResultItem import ResultItem
from ulauncher.utils.image_loader import load_image
from ulauncher.config import get_data_file
from ulauncher.search.QueryHistoryDb import QueryHistoryDb


class ShortcutResultItem(ResultItem):

    # pylint: disable=super-init-not-called, too-many-arguments
    def __init__(self, keyword, name, cmd, icon, default_search=False, run_without_argument=False, **kw):
        self.keyword = keyword
        self.name = name
        self.cmd = cmd
        self.icon = icon
        self.is_default_search = default_search
        self.run_without_argument = run_without_argument
        self._query_history = QueryHistoryDb.get_instance()

    def get_keyword(self):
        return self.keyword

    def get_name(self):
        return self.name

    def get_name_highlighted(self, query, color):
        # highlight only if we did not enter Web search item keyword
        if query.get_keyword() == self.keyword and query.get_argument():
            return None

        return super(ShortcutResultItem, self).get_name_highlighted(query, color)

    def get_description(self, query):
        if self.cmd.startswith('#!'):
            # this is a script
            description = ''
        else:
            description = self.cmd

        if self.is_default_search:
            return description.replace('%s', query)

        if query.get_keyword() == self.keyword and query.get_argument():
            return description.replace('%s', query.get_argument())
        if query.get_keyword() == self.keyword and self.run_without_argument:
            return 'Press Enter to run the shortcut'
        if query.get_keyword() == self.keyword and not query.get_argument():
            return 'Type in your query and press Enter...'

        return description.replace('%s', '...')

    def get_icon(self):
        if self.icon:
            return load_image(self.icon, self.get_icon_size())

        return load_image(get_data_file('media', 'executable-icon.png'), self.get_icon_size())

    def selected_by_default(self, query):
        """
        :param ~ulauncher.search.Query.Query query:
        """
        return self._query_history.find(query) == self.get_name()

    def on_enter(self, query):
        action_list = ActionList()
        if query.get_keyword() == self.keyword and query.get_argument():
            argument = query.get_argument()
        elif self.is_default_search:
            argument = query
        else:
            argument = None

        if self.run_without_argument:
            if self._is_url():
                action = OpenUrlAction(self.cmd.strip())
            else:
                action = RunScriptAction(self.cmd)
            action_list.append(action)
        elif argument:
            if self._is_url():
                command = self.cmd.strip().replace('%s', argument)
                action = OpenUrlAction(command)
            else:
                action = RunScriptAction(self.cmd, argument)
            action_list.append(action)
        else:
            action_list.append(SetUserQueryAction('%s ' % self.keyword))

        self._query_history.save_query(query, self.get_name())

        return action_list

    def _is_url(self) -> bool:
        return bool(re.match(r'^http(s)?://', self.cmd.strip()))



import os
from xdg.BaseDirectory import xdg_config_home

from ulauncher.search.apps.AppDb import AppDb
from ulauncher.utils.desktop.DesktopParser import DesktopParser


class AutostartPreference:
    AUTOSTART_FLAG = 'X-GNOME-Autostart-enabled'

    _ulauncher_desktop = None  # path to ulauncher.desktop
    _ulauncher_autostart_desktop = None  # path ~/.config/autostart/ulauncher.desktop

    def __init__(self):
        self._ulauncher_desktop = self._get_app_desktop()
        self._ulauncher_autostart_desktop = os.path.join(xdg_config_home, 'autostart', 'ulauncher.desktop')

    def _get_app_desktop(self):
        """
        :rtype: str
        :returns: path to desktop file
        """
        record = AppDb.get_instance().get_by_name('Ulauncher')
        if record:
            return record['desktop_file']

        return None

    def _get_autostart_parser(self):
        """
        Read ulauncher.desktop
        """
        return DesktopParser(self._ulauncher_autostart_desktop)

    def is_allowed(self):
        """
        :returns: True if autostart is allowed for Ulauncher
        """
        return bool(self._ulauncher_desktop)

    def is_on(self):
        """
        :returns: True if Ulauncher starts automatically
        """
        try:
            return self._get_autostart_parser().get_boolean(self.AUTOSTART_FLAG)
        # pylint: disable=broad-except
        except Exception:
            return False

    def switch(self, is_on):
        """
        if `is_on` is True, set `X-GNOME-Autostart-enabled=true` and
        write file to `~/.config/autostart/ulauncher.desktop`

        :param bool is_on:
        :raises SwitchError: if something goes wrong
        """
        if not self.is_allowed():
            raise SwitchError('Autostart is not allowed')

        try:
            try:
                autostart_info = self._get_autostart_parser()
            except IOError:
                autostart_info = DesktopParser(self._ulauncher_desktop)
                autostart_info.set_filename(self._ulauncher_autostart_desktop)

            autostart_info.set(self.AUTOSTART_FLAG, str(bool(is_on)).lower())
            autostart_info.set('Exec', '%s %s' % (autostart_info.get('Exec'), '--hide-window'))
            autostart_info.write()
        except Exception as e:
            raise SwitchError('Unexpected exception: %s' % e)


class SwitchError(RuntimeError):
    pass



import os
import logging
from collections import OrderedDict
from itertools import chain
from typing import Generator, List

from gi.repository import Gio

from ulauncher.utils.file_finder import find_files
from ulauncher.config import DESKTOP_DIRS, CACHE_DIR
from ulauncher.utils.Settings import Settings
from ulauncher.utils.db.KeyValueDb import KeyValueDb

logger = logging.getLogger(__name__)


def find_desktop_files(dirs: List[str] = None, pattern: str = "*.desktop") -> Generator[str, None, None]:
    """
    Returns deduped list of .desktop files (full paths)

    :param list dirs:
    :rtype: list
    """

    if dirs is None:
        dirs = DESKTOP_DIRS

    # pylint: disable=cell-var-from-loop
    all_files = chain.from_iterable(
        map(lambda f: os.path.join(f_path, f), find_files(f_path, pattern)) for f_path in dirs)

    # dedup desktop file according to follow XDG data dir order
    # specifically the first file name (i.e. firefox.desktop) take precedence
    # and other files with the same name should be ignored
    deduped_file_dict = OrderedDict()  # type: OrderedDict
    for file_path in all_files:
        file_name = os.path.basename(file_path)
        if file_name not in deduped_file_dict:
            deduped_file_dict[file_name] = file_path
    deduped_files = deduped_file_dict.values()

    blacklisted_dirs_srt = Settings.get_instance().get_property('blacklisted-desktop-dirs')
    blacklisted_dirs = blacklisted_dirs_srt.split(':') if blacklisted_dirs_srt else []
    for file in deduped_files:
        try:
            if any([file.startswith(dir) for dir in blacklisted_dirs]):
                continue
        except UnicodeDecodeError:
            continue

        yield file


def filter_app(app):
    """
    :param Gio.DesktopAppInfo app:
    :returns: True if app can be added to the database
    """
    return app and not (app.get_is_hidden() or app.get_nodisplay() or app.get_string('Type') != 'Application' or
                        not app.get_string('Name'))


def read_desktop_file(file):
    """
    :param str file: path to .desktop
    :rtype: :class:`Gio.DesktopAppInfo` or :code:`None`
    """
    try:
        return Gio.DesktopAppInfo.new_from_filename(file)
    # pylint: disable=broad-except
    except Exception as e:
        logger.info('Could not read "%s": %s', file, e)
        return None


def find_apps(dirs=None):
    """
    :param list dirs: list of paths to `*.desktop` files
    :returns: list of :class:`Gio.DesktopAppInfo` objects
    """
    if dirs is None:
        dirs = DESKTOP_DIRS

    return list(filter(filter_app, map(read_desktop_file, find_desktop_files(dirs))))


def find_apps_cached(dirs=None):
    """
    :param list dirs: list of paths to `*.desktop` files
    :returns: list of :class:`Gio.DesktopAppInfo` objects

    Pseudo code:
    >>> if cache hit:
    >>>     take list of paths from cache
    >>>     yield from filter(filter_app, map(read_desktop_file, cached_paths))
    >>> yield from find_apps()
    >>> save new paths to the cache
    """
    if dirs is None:
        dirs = DESKTOP_DIRS

    desktop_file_cache_dir = os.path.join(CACHE_DIR, 'desktop_dirs.db')
    cache = KeyValueDb(desktop_file_cache_dir).open()
    desktop_dirs = cache.find('desktop_dirs')
    if desktop_dirs:
        for dir in desktop_dirs:
            app_info = read_desktop_file(dir)
            if filter_app(app_info):
                yield app_info
        logger.info('Found %s apps in cache', len(desktop_dirs))
    new_desktop_dirs = []
    for app_info in find_apps(DESKTOP_DIRS):
        yield app_info
        new_desktop_dirs.append(app_info.get_filename())
    cache.put('desktop_dirs', new_desktop_dirs)
    cache.commit()
    logger.info('Found %s apps in the system', len(new_desktop_dirs))



import logging
from functools import partial
from threading import Timer
from gi.repository import GLib

from ulauncher.api.shared.action.DoNothingAction import DoNothingAction
from ulauncher.api.shared.action.RenderResultListAction import RenderResultListAction
from ulauncher.api.shared.item.ResultItem import ResultItem
from ulauncher.utils.decorator.singleton import singleton

logger = logging.getLogger('DeferredResultRenderer')


class DeferredResultRenderer:
    """
    Handles asynchronous render for extensions
    """

    LOADING_DELAY = 0.3  # delay in sec before Loading... is rendered

    @classmethod
    @singleton
    def get_instance(cls) -> 'DeferredResultRenderer':
        """
        Returns singleton instance
        """
        return cls()

    def __init__(self):
        self.loading = None
        self.active_event = None
        self.active_controller = None

    def get_active_controller(self):
        return self.active_controller

    def handle_event(self, event, controller):
        """
        Schedules "Loading..." message

        :rtype: :class:`~ulauncher.api.shared.action.DoNothingAction.DoNothingAction`
        """
        self._cancel_loading()
        self.loading = Timer(self.LOADING_DELAY,
                             partial(self._render_loading,
                                     controller.get_manifest().load_icon(ResultItem.get_icon_size())))
        self.loading.start()
        self.active_event = event
        self.active_controller = controller

        return DoNothingAction()

    def handle_response(self, response, controller):
        """
        Calls :func:`response.action.run`
        """
        if self.active_controller != controller or self.active_event != response.event:
            return

        self._cancel_loading()
        response.action.run()
        if not response.action.keep_app_open():
            self._hide_window()

    def on_query_change(self):
        """
        Cancel "Loading...", reset active_event and active_controller
        """
        self._cancel_loading()
        self.active_event = None
        self.active_controller = None

    def _cancel_loading(self):
        if self.loading:
            self.loading.cancel()
            self.loading = None

    def _render_loading(self, icon):
        loading_item = ResultItem(name='Loading...',
                                  icon=icon,
                                  highlightable=False)
        RenderResultListAction([loading_item]).run()

    def _hide_window(self):
        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow

        window = UlauncherWindow.get_instance()
        if window.is_visible():
            # update UI in the main thread to avoid race conditions
            GLib.idle_add(window.hide_and_clear_input)



from ulauncher.utils.SortedCollection import SortedCollection
from ulauncher.utils.fuzzy_search import get_score


class SortedList:
    """
    List maintains result items in a sorted order
    (sorted by a score, which is a similarity between item's name and a query)
    and limited to a number `limit` passed into the constructor
    """

    def __init__(self, query, min_score=30, limit=9):
        self._query = query.lower().strip()
        self._min_score = min_score
        self._limit = limit
        self._items = SortedCollection(key=lambda i: i.score)

    def __len__(self):
        return len(self._items)

    def __getitem__(self, i):
        return self._items[i]

    def __iter__(self):
        return iter(self._items)

    def __reversed__(self):
        return reversed(self._items)

    def __repr__(self):
        return '%s(%r, min_score=%s, limit=%s)' % (
            self.__class__.__name__,
            self._items,
            self._min_score,
            self._limit
        )

    def __contains__(self, item):
        return item in self._items

    def extend(self, items):
        for item in items:
            self.append(item)

    def append(self, result_item):
        score = get_score(self._query, result_item.get_search_name())
        if score >= self._min_score:
            result_item.score = -score  # use negative to sort by score in desc. order
            self._items.insert(result_item)
            while len(self._items) > self._limit:
                self._items.pop()  # remove items with the lowest score to maintain limited number of items



import os
import logging
from zipfile import ZipFile
from urllib.request import urlretrieve
from tempfile import mktemp, mkdtemp
from shutil import rmtree, move
from datetime import datetime
from ulauncher.utils.mypy_extensions import TypedDict

from ulauncher.config import EXTENSIONS_DIR
from ulauncher.utils.decorator.run_async import run_async
from ulauncher.utils.decorator.singleton import singleton
from ulauncher.api.shared.errors import UlauncherAPIError, ErrorName
from ulauncher.api.server.ExtensionDb import ExtensionDb, ExtensionRecord
from ulauncher.api.server.GithubExtension import GithubExtension
from ulauncher.api.server.ExtensionRunner import ExtensionRunner, ExtensionIsNotRunningError
from ulauncher.api.server.extension_finder import find_extensions


logger = logging.getLogger(__name__)


class ExtensionDownloaderError(UlauncherAPIError):
    pass


class ExtensionIsUpToDateError(Exception):
    pass


LastCommit = TypedDict('LastCommit', {
    'last_commit': str,
    'last_commit_time': str
})


class ExtensionDownloader:

    @classmethod
    @singleton
    def get_instance(cls) -> 'ExtensionDownloader':
        ext_db = ExtensionDb.get_instance()
        ext_runner = ExtensionRunner.get_instance()
        return cls(ext_db, ext_runner)

    def __init__(self, ext_db: ExtensionDb, ext_runner: ExtensionRunner):
        super(ExtensionDownloader, self).__init__()
        self.ext_db = ext_db
        self.ext_runner = ext_runner

    def download(self, url: str) -> str:
        """
        1. check if ext already exists
        2. get last commit info
        3. download & unzip
        4. add it to the db

        :rtype: str
        :returns: Extension ID
        :raises AlreadyDownloadedError:
        """
        gh_ext = GithubExtension(url)
        gh_ext.validate_url()

        # 1. check if ext already exists
        ext_id = gh_ext.get_ext_id()
        ext_path = os.path.join(EXTENSIONS_DIR, ext_id)
        # allow user to re-download an extension if it's not running
        # most likely it has some problems with manifest file if it's not running
        if os.path.exists(ext_path) and self.ext_runner.is_running(ext_id):
            raise ExtensionDownloaderError('Extension with URL "%s" is already added' % url,
                                           ErrorName.ExtensionAlreadyAdded)

        # 2. get last commit info
        commit = gh_ext.find_compatible_version()

        # 3. download & unzip
        filename = download_zip(gh_ext.get_download_url(commit['sha']))
        unzip(filename, ext_path)

        # 4. add to the db
        self.ext_db.put(ext_id, {
            'id': ext_id,
            'url': url,
            'updated_at': datetime.now().isoformat(),
            'last_commit': commit['sha'],
            'last_commit_time': commit['time'].isoformat()
        })
        self.ext_db.commit()

        return ext_id

    @run_async(daemon=True)
    def download_missing(self) -> None:
        already_downloaded = {id for id, _ in find_extensions(EXTENSIONS_DIR)}
        for id, ext in self.ext_db.get_records().items():
            if id in already_downloaded:
                continue

            logger.info('Downloading missing extension %s', id)
            try:
                ext_id = self.download(ext['url'])
                self.ext_runner.run(ext_id)
            # pylint: disable=broad-except
            except Exception as e:
                logger.error('%s: %s', type(e).__name__, e)

    def remove(self, ext_id: str) -> None:
        try:
            self.ext_runner.stop(ext_id)
        except ExtensionIsNotRunningError:
            pass

        rmtree(os.path.join(EXTENSIONS_DIR, ext_id))
        self.ext_db.remove(ext_id)
        self.ext_db.commit()

    def update(self, ext_id: str) -> bool:
        """
        :raises ExtensionDownloaderError:
        :rtype: boolean
        :returns: False if already up-to-date, True if was updated
        """
        commit = self.get_new_version(ext_id)
        ext = self._find_extension(ext_id)

        logger.info('Updating extension "%s" from commit %s to %s', ext_id,
                    ext['last_commit'][:8], commit['last_commit'][:8])

        if self.ext_runner.is_running(ext_id):
            self.ext_runner.stop(ext_id)

        ext_path = os.path.join(EXTENSIONS_DIR, ext_id)

        gh_ext = GithubExtension(ext['url'])
        filename = download_zip(gh_ext.get_download_url(commit['last_commit']))
        unzip(filename, ext_path)

        ext['updated_at'] = datetime.now().isoformat()
        ext['last_commit'] = commit['last_commit']
        ext['last_commit_time'] = commit['last_commit_time']
        self.ext_db.put(ext_id, ext)
        self.ext_db.commit()

        self.ext_runner.run(ext_id)

        return True

    def get_new_version(self, ext_id: str) -> LastCommit:
        """
        Returns dict with commit info about a new version or raises ExtensionIsUpToDateError
        """
        ext = self._find_extension(ext_id)
        url = ext['url']
        gh_ext = GithubExtension(url)
        commit = gh_ext.find_compatible_version()
        need_update = ext['last_commit'] != commit['sha']
        if not need_update:
            raise ExtensionIsUpToDateError('Extension is up to date')

        return {
            'last_commit': commit['sha'],
            'last_commit_time': commit['time'].isoformat()
        }

    def _find_extension(self, ext_id: str) -> ExtensionRecord:
        ext = self.ext_db.find(ext_id)
        if not ext:
            raise ExtensionDownloaderError("Extension not found", ErrorName.UnexpectedError)
        return ext


def unzip(filename: str, ext_path: str) -> None:
    """
    1. Remove ext_path
    2. Extract zip into temp dir
    3. Move contents of <temp_dir>/<project_name>-master/* to ext_path
    """
    if os.path.exists(ext_path):
        rmtree(ext_path)

    temp_ext_path = mkdtemp(prefix='ulauncher_dl_')

    with ZipFile(filename) as zipfile:
        zipfile.extractall(temp_ext_path)

    for dir in os.listdir(temp_ext_path):
        move(os.path.join(temp_ext_path, dir), ext_path)
        # there is only one directory here, so return immediately
        return


def download_zip(url: str) -> str:
    dest_zip = mktemp('.zip', prefix='ulauncher_dl_')
    filename, _ = urlretrieve(url, dest_zip)

    return filename



import os
from fnmatch import fnmatch
from typing import Callable, Generator


def find_files(directory: str, pattern: str = None, filter_fn: Callable = None) -> Generator[str, None, None]:
    """
    Search files in `directory`
    `filter_fn` takes two arguments: directory, filename.
    If return value is False, file will be ignored
    """
    for root, _, files in os.walk(directory):
        for basename in files:
            if (not pattern or fnmatch(basename, pattern)) and (not filter_fn or filter_fn(root, basename)):
                yield os.path.join(root, basename)



import operator
from functools import lru_cache
# pylint: disable=no-name-in-module
from Levenshtein import ratio


@lru_cache(maxsize=150)
def get_matching_indexes(query, text):
    """
    Uses Longest Common Substring Algorithm to find the best match

    Runs in O(nm)

    :returns: a list of positions of chars from query inside text
    """
    query = query.lower()
    text = text.lower()
    m = len(query)
    n = len(text)
    counter = [[0] * (n + 1) for x in range(m + 1)]
    for i in range(m):
        for j in range(n):
            if query[i] == text[j]:
                c = counter[i][j] + 1
                counter[i + 1][j + 1] = c

    #       F  i  r  e  f  o  x     W  e  b     B  r  o  w  s  e  r
    #  [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    # f [0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    # i [0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    # w [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
    # e [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 1, 0],
    # b [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 1, 0, 0, 0, 0, 0, 0],
    # r [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 1],
    # o [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0]]

    positions = set()
    i = m
    while i > 0:
        j, c = max(enumerate(counter[i]), key=operator.itemgetter(1))
        if c:
            for item in range(j - c, j):
                positions.add(item)
            i -= c
        else:
            i -= 1

    return sorted(positions)


def get_score(query, text):
    """
    Uses Levenshtein's algorithm + some improvements to the score
    :returns: number between 0 and 100
    """
    if not query or not text:
        return 0

    query = query.lower()
    text = text.lower()
    score = ratio(query, text) * 100

    # increase score if a word from text starts with a query
    for text_part in text.split(' '):
        if text_part.startswith(query):
            score += 30
            break

    # increase score if each separate group in indexes is a beginning of a word in text
    # example for query 'fiwebr' groups 'fi', 'we', and 'br' are matching word beginnings
    # of text 'Firefox Web Browser'
    # increase score for each such group
    increment = 10
    i = 0  # query iterator
    lq = len(query)
    for j, char in enumerate(text):
        # run until query ends and check if a query char. equals to current text char
        if i < lq and query[i] == char:
            # if char from query matches beginning of text or beginning of a word inside the text, increase the score
            if j == 0 or text[j - 1] in ' .(-_+)':
                score += increment
            i += 1
        elif i == lq:
            break

    return min(100, score)



import logging
import pickle

from ulauncher.utils.SimpleWebSocketServer import WebSocket
from ulauncher.utils.decorator.debounce import debounce
from ulauncher.api.shared.Response import Response
from ulauncher.api.shared.event import KeywordQueryEvent, PreferencesEvent, PreferencesUpdateEvent
from ulauncher.api.server.DeferredResultRenderer import DeferredResultRenderer
from ulauncher.api.server.ExtensionPreferences import ExtensionPreferences
from ulauncher.api.server.ExtensionManifest import ExtensionManifest, ExtensionManifestError

logger = logging.getLogger(__name__)


class ExtensionController(WebSocket):
    """
    Handles communication between Ulauncher app and an extension.

    :param list controllers: list of :class:`~ulauncher.api.server.ExtensionController`
    """

    extension_id = None
    manifest = None
    preferences = None
    _debounced_send_event = None

    def __init__(self, controllers, *args, **kw):
        self.controllers = controllers
        self.result_renderer = DeferredResultRenderer.get_instance()  # type: DeferredResultRenderer
        super(ExtensionController, self).__init__(*args, **kw)

    def _send_event(self, event):
        logger.debug('Send event %s to "%s"', type(event).__name__, self.extension_id)
        self.sendMessage(pickle.dumps(event))

    def handle_query(self, query):
        """
        Handles user query with a keyword from this extension

        :returns: :class:`BaseAction` object
        """
        event = KeywordQueryEvent(query)
        return self.trigger_event(event)

    def trigger_event(self, event):
        """
        Triggers event for an extension
        """
        # don't debounce events that are triggered by updates in preferences
        if isinstance(event, PreferencesUpdateEvent):
            self._send_event(event)
        else:
            self._debounced_send_event(event)
        return self.result_renderer.handle_event(event, self)

    def get_manifest(self):
        return self.manifest

    def get_extension_id(self):
        return self.extension_id

    def handleMessage(self):
        """
        Implements abstract method of :class:`WebSocket`

        Handles incoming message stored in `self.data` by invoking
        :meth:`~ulauncher.api.server.DeferredResultRenderer.DeferredResultRenderer.handle_response`
        of `DeferredResultRenderer`
        """
        response = pickle.loads(self.data)

        if not isinstance(response, Response):
            raise Exception("Unsupported type %s" % type(response).__name__)

        logger.debug('Incoming response (%s, %s) from "%s"', type(response.event).__name__,
                     type(response.action).__name__,
                     self.extension_id)
        self.result_renderer.handle_response(response, self)

    def handleConnected(self):
        """
        Implements abstract method of :class:`WebSocket`

        * Appends `self` to `self.controllers`
        * Validates manifest file.
        * Sends :class:`PreferencesEvent` to extension
        """
        self.extension_id = self.request.path[1:]
        if not self.extension_id:
            raise Exception('Incorrect path %s' % self.request.path)

        logger.info('Extension "%s" connected', self.extension_id)

        self.manifest = ExtensionManifest.open(self.extension_id)
        try:
            self.manifest.validate()
        except ExtensionManifestError as e:
            logger.warning("Couldn't connect '%s'. %s: %s", self.extension_id, type(e).__name__, e)
            self.close()
            return

        self.preferences = ExtensionPreferences.create_instance(self.extension_id)
        self.controllers[self.extension_id] = self
        self._debounced_send_event = debounce(self.manifest.get_option('query_debounce', 0.05))(self._send_event)

        self._send_event(PreferencesEvent(self.preferences.get_dict()))

    def handleClose(self):
        """
        Implements abstract method of :class:`WebSocket`
        """
        logger.info('Extension "%s" disconnected', self.extension_id)
        try:
            del self.controllers[self.extension_id]
        # pylint: disable=broad-except
        except Exception:
            pass



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: Bind a new key
# Where your project will look for your data (for instance, images and ui
# files). By default, this is ../data, relative your trunk layout
# pylint: disable=deprecated-module
import optparse

__ulauncher_data_directory__ = '/usr/share/ulauncher/'
__license__ = 'GPL-3'
__version__ = '5.7.0'

import os
from uuid import uuid4
from time import time
from functools import lru_cache

from gettext import gettext
from xdg.BaseDirectory import xdg_config_home, xdg_cache_home, xdg_data_dirs, xdg_data_home

DATA_DIR = os.path.join(xdg_data_home, 'ulauncher')
# Use ulauncher_cache dir because of the WebKit bug
# https://bugs.webkit.org/show_bug.cgi?id=151646
CACHE_DIR = os.path.join(xdg_cache_home, 'ulauncher_cache')
CONFIG_DIR = os.path.join(xdg_config_home, 'ulauncher')
SETTINGS_FILE_PATH = os.path.join(CONFIG_DIR, 'settings.json')
# spec: https://specifications.freedesktop.org/menu-spec/latest/ar01s02.html
DESKTOP_DIRS = list(filter(os.path.exists, [os.path.join(dir, "applications") for dir in xdg_data_dirs]))
EXTENSIONS_DIR = os.path.join(DATA_DIR, 'extensions')
EXT_PREFERENCES_DIR = os.path.join(CONFIG_DIR, 'ext_preferences')
ULAUNCHER_APP_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))


class ProjectPathNotFoundError(Exception):
    """Raised when we can't find the project directory."""


def get_data_file(*path_segments):
    """Get the full path to a data file.

    :returns: path to a file underneath the data directory (as defined by :func:`get_data_path`).
              Equivalent to :code:`os.path.join(get_data_path(),*path_segments)`.
    """
    return os.path.join(get_data_path(), *path_segments)


def get_data_path():
    """Retrieve ulauncher data path

    This path is by default `<ulauncher_path>/../data/` in trunk
    and `/usr/share/ulauncher` in an installed version but this path
    is specified at installation time.
    """

    # Get pathname absolute or relative.
    path = os.path.join(
        os.path.dirname(__file__), __ulauncher_data_directory__)

    abs_data_path = os.path.abspath(path)
    if not os.path.exists(abs_data_path):
        raise ProjectPathNotFoundError(abs_data_path)

    return abs_data_path


def is_wayland():
    return os.environ.get('XDG_SESSION_TYPE', '').lower() == 'wayland'


def is_wayland_compatibility_on():
    """
    In this mode user won't be able to set app hotkey via preferences
    Set hotkey in OS Settings > Devices > Keyboard > Add Hotkey > Command: ulauncher-toggle
    """
    return is_wayland() and gdk_backend().lower() == 'wayland'


def gdk_backend():
    return os.environ.get('GDK_BACKEND', '')


@lru_cache()
def get_options():
    """Support for command line options"""
    parser = optparse.OptionParser(version="%%prog %s" % get_version())
    parser.add_option(
        "-v", "--verbose", action="count", dest="verbose",
        help=gettext("Show debug messages"))
    parser.add_option(
        "--hide-window", action="store_true",
        help=gettext("Hide window upon application startup"))
    parser.add_option(
        "--no-extensions", action="store_true",
        help=gettext("Do not run extensions"))
    parser.add_option(
        "--no-window-shadow", action="store_true",
        help=gettext("Removes window shadow. On DEs without a compositor this solves issue with a black border"))
    parser.add_option(
        "--dev", action="store_true",
        help=gettext("Enables context menu in the Preferences UI"))
    (options, _) = parser.parse_args()

    return options


def get_default_shortcuts():
    google = {
        "id": str(uuid4()),
        "name": "Google Search",
        "keyword": "g",
        "cmd": "https://google.com/search?q=%s",
        "icon": get_data_file('media/google-search-icon.png'),
        "is_default_search": True,
        "run_without_argument": False,
        "added": time()
    }
    stackoverflow = {
        "id": str(uuid4()),
        "name": "Stack Overflow",
        "keyword": "so",
        "cmd": "http://stackoverflow.com/search?q=%s",
        "icon": get_data_file('media/stackoverflow-icon.svg'),
        "is_default_search": True,
        "run_without_argument": False,
        "added": time()
    }
    wikipedia = {
        "id": str(uuid4()),
        "name": "Wikipedia",
        "keyword": "wiki",
        "cmd": "https://en.wikipedia.org/wiki/%s",
        "icon": get_data_file('media/wikipedia-icon.png'),
        "is_default_search": True,
        "run_without_argument": False,
        "added": time()
    }

    return {
        google['id']: google,
        stackoverflow['id']: stackoverflow,
        wikipedia['id']: wikipedia,
    }


def get_version():
    return __version__



import os
from ulauncher.config import DATA_DIR
from ulauncher.utils.db.KeyValueDb import KeyValueDb
from ulauncher.utils.decorator.singleton import singleton


class QueryHistoryDb(KeyValueDb):

    @classmethod
    @singleton
    def get_instance(cls):
        db = cls(os.path.join(DATA_DIR, 'query_history.db'))
        db.open()
        return db

    def save_query(self, query: str, item_name: str):
        """
        Saves "query -> item_name" mapping to history
        so later ulauncher can autoselect item with that name
        if the same query is entered
        """
        if not query:
            # ignore cases when item is selected from the list of frequent apps
            return

        self.put(query, item_name)
        self.commit()



from time import sleep
from gi.repository import GLib

from ulauncher.api.shared.action.BaseAction import BaseAction


class SetUserQueryAction(BaseAction):
    """
    Changes query string to a new one

    :param str new_query:
    """

    def __init__(self, new_query):
        self.new_query = new_query

    def keep_app_open(self):
        return True

    def run(self):
        GLib.idle_add(self._update_query)

    def _update_query(self):
        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow

        input = UlauncherWindow.get_instance().get_input()
        input.set_text(self.new_query)

        # Ugly hack:
        # Defer set position, because GTK sets position after change event occurs
        sleep(0.002)
        input.set_position(len(self.new_query))



from ulauncher.api.shared.action.BaseAction import BaseAction


class HideWindowAction(BaseAction):
    """
    Does what the class name says
    """

    def keep_app_open(self):
        return False

    def run(self):
        pass



import os
import sys
import logging
from random import randint

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"


def setup_logging():
    root = logging.getLogger()

    ext_name = COLOR_SEQ % (30 + randint(1, 8)) + get_extension_name() + RESET_SEQ
    formatter = logging.Formatter(
        ext_name + " | %(asctime)s | %(levelname)s | %(name)s: %(funcName)s() | %(message)s")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    root.addHandler(handler)
    root.setLevel(logging.WARNING)

    if os.getenv('VERBOSE'):
        root.setLevel(logging.DEBUG)


def get_extension_name():
    return os.path.basename(os.path.dirname(sys.argv[0]))



import os
import logging
from typing import List, Optional, Dict
from functools import lru_cache
from ulauncher.config import EXT_PREFERENCES_DIR
from ulauncher.utils.db.KeyValueDb import KeyValueDb
from ulauncher.utils.mypy_extensions import TypedDict
from ulauncher.api.server.ExtensionManifest import ExtensionManifest, OptionItems

logger = logging.getLogger(__name__)

PreferenceItem = TypedDict('PreferenceItem', {
    'id': str,
    'type': str,
    'name': str,
    'description': str,
    'options': OptionItems,
    'default_value': str,
    'user_value': str,
    'value': str,
})
PreferenceItems = List[PreferenceItem]


class ExtensionPreferences:
    """
    Manages extension preferences. Stores them in pickled file in cache directory
    """

    manifest = None  # type: ExtensionManifest

    @classmethod
    @lru_cache(maxsize=1000)
    def create_instance(cls, ext_id):
        return cls(ext_id, ExtensionManifest.open(ext_id))

    def __init__(self, ext_id: str, manifest: ExtensionManifest, ext_preferences_dir: str = EXT_PREFERENCES_DIR):
        self.db_path = os.path.join(ext_preferences_dir, '%s.db' % ext_id)
        self.db = KeyValueDb[str, str](self.db_path)
        self.manifest = manifest
        self._db_is_open = False

    def get_items(self, type: str = None) -> PreferenceItems:
        """
        :param str type:
        :rtype: list of dicts: [{id: .., type: .., defalut_value: .., user_value: ..., value: ..., description}, ...]
        """
        self._open_db()

        items = []  # type: PreferenceItems
        for p in self.manifest.get_preferences():
            if type and type != p['type']:
                continue

            default_value = p.get('default_value', '')
            items.append({
                'id': p['id'],
                'type': p['type'],
                'name': p.get('name', ''),
                'description': p.get('description', ''),
                'options': p.get('options', []),
                'default_value': default_value,
                'user_value': self.db.find(p['id']) or '',
                'value': self.db.find(p['id']) or default_value
            })

        return items

    def get_dict(self) -> Dict[str, str]:
        """
        :rtype: dict(id=value, id2=value2, ...)
        """
        items = {}
        for i in self.get_items():
            items[i['id']] = i['value']

        return items

    def get(self, id: str) -> Optional[PreferenceItem]:
        """
        Returns one item
        """
        for i in self.get_items():
            if i['id'] == id:
                return i

        return None

    def get_active_keywords(self) -> List[str]:
        """
        Filters items by type "keyword"
        """
        return [p['value'] for p in self.get_items(type='keyword') if p['value']]

    def set(self, id: str, value: str):
        """
        Updates preference

        :param str id: id as defined in manifest
        :param str value:
        """
        self.db.put(id, value)
        self.db.commit()

    def _open_db(self):
        if not self._db_is_open:
            self.db.open()
            self._db_is_open = True



import sys
import os
import signal
import logging
import time
from threading import Event
import gi

# Fixes issue #488
sys.path.append('/usr/lib/python3.8/site-packages')

gi.require_version('Gtk', '3.0')
# pylint: disable=wrong-import-position
from gi.repository import Gtk
import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop

from ulauncher.config import (get_version, get_options, is_wayland, is_wayland_compatibility_on,
                              gdk_backend, CACHE_DIR, CONFIG_DIR, DATA_DIR)
from ulauncher.utils.decorator.run_async import run_async
from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow
from ulauncher.ui.AppIndicator import AppIndicator
from ulauncher.utils.Settings import Settings
from ulauncher.utils.setup_logging import setup_logging
from ulauncher.api.version import api_version


DBUS_SERVICE = 'net.launchpad.ulauncher'
DBUS_PATH = '/net/launchpad/ulauncher'


def _create_dirs():
    if not os.path.exists(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)

    if not os.path.exists(DATA_DIR):
        os.makedirs(DATA_DIR)

    # make sure ~/.cache/ulauncher exists
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)


class UlauncherDbusService(dbus.service.Object):
    def __init__(self, window):
        self.window = window
        bus_name = dbus.service.BusName(DBUS_SERVICE, bus=dbus.SessionBus())
        super(UlauncherDbusService, self).__init__(bus_name, DBUS_PATH)

    @dbus.service.method(DBUS_SERVICE)
    def toggle_window(self):
        self.window.toggle_window()


# pylint: disable=too-few-public-methods
class SignalHandler:

    _exit_event = None
    _app_window = None
    _logger = None

    def __init__(self, app_window):
        self._exit_event = Event()
        self._app_window = app_window
        self._logger = logging.getLogger('ulauncher')
        signal.signal(signal.SIGINT, self._exit_gracefully)
        signal.signal(signal.SIGTERM, self._exit_gracefully)
        signal.signal(signal.SIGHUP, self._reload_configs)

    def _reload_configs(self, *args):
        self._logger.info('Received SIGHUP. Reloading configs')
        self._app_window.init_theme()

    def killed(self):
        """
        :rtype: bool
        """
        return self._exit_event.is_set()

    def _exit_gracefully(self, *args):
        self._exit_event.set()


def main():
    """
    Main function that starts everything
    """
    if is_wayland() and gdk_backend().lower() != 'x11' and not is_wayland_compatibility_on():
        warn = """
                    [!]
        Looks like you are in Wayland session
        Please run Ulauncher with env var
        GDK_BACKEND set to 'x11' like this:

        GDK_BACKEND=x11 ulauncher
        """
        print(warn, file=sys.stderr)
        sys.exit(1)

    # start DBus loop
    DBusGMainLoop(set_as_default=True)
    bus = dbus.SessionBus()
    instance = bus.request_name(DBUS_SERVICE)

    if instance != dbus.bus.REQUEST_NAME_REPLY_PRIMARY_OWNER:
        toggle_window = dbus.SessionBus().get_object(DBUS_SERVICE, DBUS_PATH).get_dbus_method("toggle_window")
        toggle_window()
        return

    _create_dirs()

    options = get_options()
    setup_logging(options)
    logger = logging.getLogger('ulauncher')
    logger.info('Ulauncher version %s', get_version())
    logger.info('Extension API version %s', api_version)
    logger.info("GTK+ %s.%s.%s", Gtk.get_major_version(), Gtk.get_minor_version(), Gtk.get_micro_version())
    logger.info("Is Wayland: %s", is_wayland())
    logger.info("Wayland compatibility: %s", ('on' if is_wayland_compatibility_on() else 'off'))

    # log uncaught exceptions
    def except_hook(exctype, value, tb):
        logger.error("Uncaught exception", exc_info=(exctype, value, tb))

    sys.excepthook = except_hook

    window = UlauncherWindow.get_instance()
    UlauncherDbusService(window)
    if not options.hide_window:
        window.show()

    if Settings.get_instance().get_property('show-indicator-icon'):
        AppIndicator.get_instance().show()

    # workaround to make Ctrl+C quiting the app
    signal_handler = SignalHandler(window)
    gtk_thread = run_async(Gtk.main)()
    try:
        while gtk_thread.is_alive() and not signal_handler.killed():
            time.sleep(0.5)
    except KeyboardInterrupt:
        logger.warning('On KeyboardInterrupt')
    finally:
        Gtk.main_quit()



import re
import ast
import operator as op

from ulauncher.api.shared.action.RenderResultListAction import RenderResultListAction
from ulauncher.search.BaseSearchMode import BaseSearchMode
from ulauncher.search.calc.CalcResultItem import CalcResultItem


# supported operators
operators = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
             ast.Div: op.truediv, ast.Pow: op.pow, ast.BitXor: op.xor,
             ast.USub: op.neg}


def eval_expr(expr):
    """
    >>> eval_expr('2^6')
    4
    >>> eval_expr('2**6')
    64
    >>> eval_expr('2*6+')
    12
    >>> eval_expr('1 + 2*3**(4^5) / (6 + -7)')
    -5.0
    """
    try:
        return _eval(ast.parse(expr, mode='eval').body)
    # pylint: disable=broad-except
    except Exception:
        # if failed, try without the last symbol
        return _eval(ast.parse(expr[:-1], mode='eval').body)


def _eval(node):
    if isinstance(node, ast.Num):  # <number>
        return node.n
    if isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return operators[type(node.op)](_eval(node.left), _eval(node.right))
    if isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return operators[type(node.op)](_eval(node.operand))

    raise TypeError(node)


class CalcMode(BaseSearchMode):
    RE_CALC = re.compile(r'^[\d\-\(\.][\d\*+\/\-\.e\(\)\^ ]*$', flags=re.IGNORECASE)

    def is_enabled(self, query):
        return re.match(self.RE_CALC, query)

    def handle_query(self, query):
        try:
            result = eval_expr(query)
            if result is None:
                raise ValueError()

            # fixes issue with division where result is represented as a float (e.g., 1.0)
            # although it is an integer (1)
            if int(result) == result:
                result = int(result)

            result_item = CalcResultItem(result=result)
        # pylint: disable=broad-except
        except Exception:
            result_item = CalcResultItem(error='Invalid expression')
        return RenderResultListAction([result_item])
